@@ -6,9 +6,9 @@ use std::{
 };
 
 use crc::Crc;
-use futures_util::{future, FutureExt, StreamExt, TryFutureExt, TryStreamExt};
-use quinn::{ConnectionError, ReadError, WriteError};
+use quinn::{ConnectionError, ReadError, StoppedError, TransportConfig, WriteError};
 use rand::{self, RngCore};
+use rustls::pki_types::{CertificateDer, PrivatePkcs8KeyDer};
 use tokio::runtime::Builder;
 
 struct Shared {
@@ -26,55 +26,54 @@ fn connect_n_nodes_to_1_and_send_1mb_data() {
     .unwrap();
 
     let runtime = Builder::new_current_thread().enable_all().build().unwrap();
+    let _guard = runtime.enter();
     let shared = Arc::new(Mutex::new(Shared { errors: vec![] }));
 
     let (cfg, listener_cert) = configure_listener();
-    let (endpoint, incoming_conns) = {
-        let _guard = runtime.enter();
-        quinn::Endpoint::server(cfg, "127.0.0.1:0".parse().unwrap()).unwrap()
-    };
+    let endpoint = quinn::Endpoint::server(cfg, "127.0.0.1:0".parse().unwrap()).unwrap();
     let listener_addr = endpoint.local_addr().unwrap();
 
     let expected_messages = 50;
 
     let crc = crc::Crc::<u32>::new(&crc::CRC_32_ISO_HDLC);
     let shared2 = shared.clone();
-    let read_incoming_data = incoming_conns
-        .filter_map(|connect| connect.map(|x| x.ok()))
-        .take(expected_messages)
-        .for_each(move |new_conn| {
-            let conn = new_conn.connection;
+    let endpoint2 = endpoint.clone();
+    let read_incoming_data = async move {
+        for _ in 0..expected_messages {
+            let conn = endpoint2.accept().await.unwrap().await.unwrap();
 
             let shared = shared2.clone();
-            let task = new_conn
-                .uni_streams
-                .try_for_each(move |stream| {
-                    let conn = conn.clone();
-                    read_from_peer(stream).map(move |_| {
-                        conn.close(0u32.into(), &[]);
-                        Ok(())
-                    })
-                })
-                .unwrap_or_else(move |e| {
+            let task = async move {
+                while let Ok(stream) = conn.accept_uni().await {
+                    read_from_peer(stream).await?;
+                    conn.close(0u32.into(), &[]);
+                }
+                Ok(())
+            };
+            tokio::spawn(async move {
+                if let Err(e) = task.await {
                     shared.lock().unwrap().errors.push(e);
-                });
-            tokio::spawn(task);
-
-            future::ready(())
-        });
+                }
+            });
+        }
+    };
     runtime.spawn(read_incoming_data);
 
-    let client_cfg = configure_connector(&listener_cert);
+    let client_cfg = configure_connector(listener_cert);
 
     for _ in 0..expected_messages {
         let data = random_data_with_hash(1024 * 1024, &crc);
         let shared = shared.clone();
-        let task = endpoint
+        let connecting = endpoint
             .connect_with(client_cfg.clone(), listener_addr, "localhost")
-            .unwrap()
-            .map_err(WriteError::ConnectionLost)
-            .and_then(move |new_conn| write_to_peer(new_conn.connection, data))
-            .unwrap_or_else(move |e| {
+            .unwrap();
+        let task = async move {
+            let conn = connecting.await.map_err(WriteError::ConnectionLost)?;
+            write_to_peer(conn, data).await?;
+            Ok(())
+        };
+        runtime.spawn(async move {
+            if let Err(e) = task.await {
                 use quinn::ConnectionError::*;
                 match e {
                     WriteError::ConnectionLost(ApplicationClosed { .. })
@@ -82,8 +81,8 @@ fn connect_n_nodes_to_1_and_send_1mb_data() {
                     WriteError::ConnectionLost(e) => shared.lock().unwrap().errors.push(e),
                     _ => panic!("unexpected write error"),
                 }
-            });
-        runtime.spawn(task);
+            }
+        });
     }
 
     runtime.block_on(endpoint.wait_idle());
@@ -93,7 +92,7 @@ fn connect_n_nodes_to_1_and_send_1mb_data() {
     }
 }
 
-async fn read_from_peer(stream: quinn::RecvStream) -> Result<(), quinn::ConnectionError> {
+async fn read_from_peer(mut stream: quinn::RecvStream) -> Result<(), quinn::ConnectionError> {
     let crc = crc::Crc::<u32>::new(&crc::CRC_32_ISO_HDLC);
     match stream.read_to_end(1024 * 1024 * 5).await {
         Ok(data) => {
@@ -104,11 +103,10 @@ async fn read_from_peer(stream: quinn::RecvStream) -> Result<(), quinn::Connecti
             use quinn::ReadToEndError::*;
             use ReadError::*;
             match e {
-                TooLong
-                | Read(UnknownStream)
-                | Read(ZeroRttRejected)
-                | Read(IllegalOrderedRead) => unreachable!(),
-                Read(Reset(error_code)) => panic!("unexpected stream reset: {}", error_code),
+                TooLong | Read(ClosedStream) | Read(ZeroRttRejected) | Read(IllegalOrderedRead) => {
+                    unreachable!()
+                }
+                Read(Reset(error_code)) => panic!("unexpected stream reset: {error_code}"),
                 Read(ConnectionLost(e)) => Err(e),
             }
         }
@@ -118,30 +116,33 @@ async fn read_from_peer(stream: quinn::RecvStream) -> Result<(), quinn::Connecti
 async fn write_to_peer(conn: quinn::Connection, data: Vec<u8>) -> Result<(), WriteError> {
     let mut s = conn.open_uni().await.map_err(WriteError::ConnectionLost)?;
     s.write_all(&data).await?;
-    // Suppress finish errors, since the peer may close before ACKing
-    match s.finish().await {
-        Ok(()) => Ok(()),
-        Err(WriteError::ConnectionLost(ConnectionError::ApplicationClosed { .. })) => Ok(()),
-        Err(e) => Err(e),
+    s.finish().unwrap();
+    // Wait for the stream to be fully received
+    match s.stopped().await {
+        Ok(_) => Ok(()),
+        Err(StoppedError::ConnectionLost(ConnectionError::ApplicationClosed { .. })) => Ok(()),
+        Err(e) => Err(e.into()),
     }
 }
 
 /// Builds client configuration. Trusts given node certificate.
-fn configure_connector(node_cert: &rustls::Certificate) -> quinn::ClientConfig {
+fn configure_connector(node_cert: CertificateDer<'static>) -> quinn::ClientConfig {
     let mut roots = rustls::RootCertStore::empty();
     roots.add(node_cert).unwrap();
-    let mut peer_cfg = quinn::ClientConfig::with_root_certificates(roots);
-    let transport_config = Arc::get_mut(&mut peer_cfg.transport).unwrap();
+
+    let mut transport_config = TransportConfig::default();
     transport_config.max_idle_timeout(Some(Duration::from_secs(20).try_into().unwrap()));
 
+    let mut peer_cfg = quinn::ClientConfig::with_root_certificates(Arc::new(roots)).unwrap();
+    peer_cfg.transport_config(Arc::new(transport_config));
     peer_cfg
 }
 
 /// Builds listener configuration along with its certificate.
-fn configure_listener() -> (quinn::ServerConfig, rustls::Certificate) {
+fn configure_listener() -> (quinn::ServerConfig, CertificateDer<'static>) {
     let (our_cert, our_priv_key) = gen_cert();
     let mut our_cfg =
-        quinn::ServerConfig::with_single_cert(vec![our_cert.clone()], our_priv_key).unwrap();
+        quinn::ServerConfig::with_single_cert(vec![our_cert.clone()], our_priv_key.into()).unwrap();
 
     let transport_config = Arc::get_mut(&mut our_cfg.transport).unwrap();
     transport_config.max_idle_timeout(Some(Duration::from_secs(20).try_into().unwrap()));
@@ -149,10 +150,12 @@ fn configure_listener() -> (quinn::ServerConfig, rustls::Certificate) {
     (our_cfg, our_cert)
 }
 
-fn gen_cert() -> (rustls::Certificate, rustls::PrivateKey) {
+fn gen_cert() -> (CertificateDer<'static>, PrivatePkcs8KeyDer<'static>) {
     let cert = rcgen::generate_simple_self_signed(vec!["localhost".to_string()]).unwrap();
-    let key = rustls::PrivateKey(cert.serialize_private_key_der());
-    (rustls::Certificate(cert.serialize_der().unwrap()), key)
+    (
+        cert.cert.into(),
+        PrivatePkcs8KeyDer::from(cert.key_pair.serialize_der()),
+    )
 }
 
 /// Constructs a buffer with random bytes of given size prefixed with a hash of this data.
